/*
    FreeRTOS V7.4.0 - Copyright (C) 2013 Real Time Engineers Ltd.

    FEATURES AND PORTS ARE ADDED TO FREERTOS ALL THE TIME.  PLEASE VISIT
    http://www.FreeRTOS.org TO ENSURE YOU ARE USING THE LATEST VERSION.

    ***************************************************************************
     *                                                                       *
     *    FreeRTOS tutorial books are available in pdf and paperback.        *
     *    Complete, revised, and edited pdf reference manuals are also       *
     *    available.                                                         *
     *                                                                       *
     *    Purchasing FreeRTOS documentation will not only help you, by       *
     *    ensuring you get running as quickly as possible and with an        *
     *    in-depth knowledge of how to use FreeRTOS, it will also help       *
     *    the FreeRTOS project to continue with its mission of providing     *
     *    professional grade, cross platform, de facto standard solutions    *
     *    for microcontrollers - completely free of charge!                  *
     *                                                                       *
     *    >>> See http://www.FreeRTOS.org/Documentation for details. <<<     *
     *                                                                       *
     *    Thank you for using FreeRTOS, and thank you for your support!      *
     *                                                                       *
    ***************************************************************************


    This file is part of the FreeRTOS distribution.

    FreeRTOS is free software; you can redistribute it and/or modify it under
    the terms of the GNU General Public License (version 2) as published by the
    Free Software Foundation AND MODIFIED BY the FreeRTOS exception.

    >>>>>>NOTE<<<<<< The modification to the GPL is included to allow you to
    distribute a combined work that includes FreeRTOS without being obliged to
    provide the source code for proprietary components outside of the FreeRTOS
    kernel.

    FreeRTOS is distributed in the hope that it will be useful, but WITHOUT ANY
    WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
    FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
    details. You should have received a copy of the GNU General Public License
    and the FreeRTOS license exception along with FreeRTOS; if not itcan be
    viewed here: http://www.freertos.org/a00114.html and also obtained by
    writing to Real Time Engineers Ltd., contact details for whom are available
    on the FreeRTOS WEB site.

    1 tab == 4 spaces!

    ***************************************************************************
     *                                                                       *
     *    Having a problem?  Start by reading the FAQ "My application does   *
     *    not run, what could be wrong?"                                     *
     *                                                                       *
     *    http://www.FreeRTOS.org/FAQHelp.html                               *
     *                                                                       *
    ***************************************************************************


    http://www.FreeRTOS.org - Documentation, books, training, latest versions, 
    license and Real Time Engineers Ltd. contact details.

    http://www.FreeRTOS.org/plus - A selection of FreeRTOS ecosystem products,
    including FreeRTOS+Trace - an indispensable productivity tool, and our new
    fully thread aware and reentrant UDP/IP stack.

    http://www.OpenRTOS.com - Real Time Engineers ltd license FreeRTOS to High 
    Integrity Systems, who sell the code with commercial support, 
    indemnification and middleware, under the OpenRTOS brand.
    
    http://www.SafeRTOS.com - High Integrity Systems also provide a safety 
    engineered and independently SIL3 certified version for use in safety and 
    mission critical applications that require provable dependability.
*/


/* BASIC INTERRUPT DRIVEN SERIAL PORT DRIVER. 

NOTE:  This driver is primarily to test the scheduler functionality.  It does
not effectively use the buffers or DMA and is therefore not intended to be
an example of an efficient driver. */

/* This driver is for PIC32MZ on MEB II demo board */

/* Standard include file. */
#include <stdlib.h>

/* Scheduler include files. */
#include "FreeRTOS.h"
#include "queue.h"
#include "task.h"

/* Demo app include files. */
#include "serial.h"

/* Hardware setup. */
#define serSET_FLAG						( 1 )

#if ((__PIC32MZ ))
#warning IN PIC32MZ
#else
#error No processor defined in UART2
#endif

/* The queues used to communicate between tasks and ISR's. */
static xQueueHandle xRxedChars; 
static xQueueHandle xCharsForTx; 

/* Flag used to indicate the tx status. */
static portBASE_TYPE xTxHasEnded;
char junk;
/*-----------------------------------------------------------*/

/* The UART interrupt handler.  As this uses the FreeRTOS assembly interrupt
entry point the IPL setting in the following prototype has no effect.  The
interrupt priority is set by the call to  ConfigIntUART2() in
xSerialPortInitMinimal(). */

void __attribute__( (interrupt(ipl0), vector(_UART2_RX_VECTOR)))	vU2RXInterruptWrapper( void );
void __attribute__( (interrupt(ipl0), vector(_UART2_TX_VECTOR)))	vU2TXInterruptWrapper( void );
void __attribute__( (interrupt(ipl0), vector(_UART2_FAULT_VECTOR))) vU2FAULTInterruptWrapper( void );

/*-----------------------------------------------------------*/

xComPortHandle xSerialPortInitMinimal( unsigned long ulWantedBaud, unsigned portBASE_TYPE uxQueueLength )
{
unsigned short usBRG;

	/* Create the queues used by the com test task. */
	xRxedChars  = xQueueCreate( uxQueueLength, ( unsigned portBASE_TYPE ) sizeof( signed char ) );
	xCharsForTx = xQueueCreate( uxQueueLength, ( unsigned portBASE_TYPE ) sizeof( signed char ) );

	/* Need to configuer PPS pins for UART2 */

#if defined (__PIC32MZ)

	// Note: AD1PCFG is now analog bit ANSELx, x = Port bank, 1=analog, 0-digital

	ANSELGbits.ANSG6 = 0;
	U2RXR = 1;		// RPG6 --> U2RX  Pin 11 on J2 of MEB II board
	RPB14R = 2;		// U2TX --> RPB14 Pin 19 on J2 of MEB II board

#else
#error PPS pins not defined in UART2
#endif

	/* Configure the UART and interrupts. */
	usBRG = (unsigned short)(( (float)configPERIPHERAL_CLOCK_HZ / ( (float)16 * (float)ulWantedBaud ) ) - (float)0.5);  // BRGH = 0
	// usBRG = (unsigned short)(( (float)configPERIPHERAL_CLOCK_HZ / ( (float)4 * (float)ulWantedBaud ) ) - (float)0.5);  // BRGH = 1

	U2MODEbits.BRGH = 0;	// 0 = 16X clock, 1 = 4X clock
	U2BRG = usBRG;			// Set baud rate

	U2STAbits.URXISEL = 0b10;	// RX Interrupt when buffer 3/4 full
	U2STAbits.UTXISEL = 0b00;	// TX Interrupt when buffer empty

	U2STAbits.OERR = 0;			// Clear any errors
	U2STAbits.FERR = 0;
	U2STAbits.PERR = 0;
	
	while (U2STAbits.URXDA == 1)  junk = U2RXREG;	// Clear FIFO

	U2MODEbits.LPBACK = 0;  // Loopback bit

	IEC4bits.U2RXIE = 1;	// Enable interrupts
	IEC4bits.U2EIE = 1;
    /* Turn off UART TX interrupt, we will turn it on in the Queue handler */
    IEC4bits.U2TXIE = 0;

	IPC36bits.U2RXIP = configKERNEL_INTERRUPT_PRIORITY + 1;
	IPC36bits.U2TXIP = configKERNEL_INTERRUPT_PRIORITY + 1;
	IPC36bits.U2EIP = configKERNEL_INTERRUPT_PRIORITY + 1;

	IPC36bits.U2RXIS = 1;	// sub priorty (0<<4)
	IPC36bits.U2TXIS = 0;
	IPC36bits.U2EIS = 0;

	//U2MODEbits.UEN = 0b00;	// Enable UART Pins (not needed on parts with PPS)
	U2STAbits.URXEN = 1;	// Rx Enable
	U2STAbits.UTXEN = 1;	// Tx Enable
	
	IFS4bits.U2RXIF = 0;	// Clear any interrupts that might be pending
	IFS4bits.U2TXIF = 0;
	IFS4bits.U2EIF = 0;

	U2MODEbits.ON = 1;		// UART enable

    xTxHasEnded = pdTRUE;

	/* Only a single port is implemented so we don't need to return anything. */
	return NULL;
}
/*-----------------------------------------------------------*/

signed portBASE_TYPE xSerialGetChar( xComPortHandle pxPort, signed char *pcRxedChar, portTickType xBlockTime )
{
	/* Only one port is supported. */
	( void ) pxPort;

	/* Get the next character from the buffer.  Return false if no characters
	are available or arrive before xBlockTime expires. */
	if( xQueueReceive( xRxedChars, pcRxedChar, xBlockTime ) )
	{
            return pdTRUE;
	}
	else
	{
            return pdFALSE;
	}
}
/*-----------------------------------------------------------*/

signed portBASE_TYPE xSerialPutChar( xComPortHandle pxPort, signed char cOutChar, portTickType xBlockTime )
{
	/* Only one port is supported. */
	( void ) pxPort;


	/* Return false if after the block time there is no room on the Tx queue. */
	if( xQueueSend( xCharsForTx, &cOutChar, xBlockTime ) != pdPASS )
	{
            return pdFAIL;
	}

	/* A critical section should not be required as xTxHasEnded will not be
	written to by the ISR if it is already 0 (is this correct?). */
	if( xTxHasEnded )  
	{
		
            xTxHasEnded = pdFALSE;
            /* We need to enable UART TX interrupt */
			IEC4bits.U2TXIE = 1;
	}

	return pdPASS;
}
/*-----------------------------------------------------------*/

void vSerialClose( xComPortHandle xPort )
{
/* Only one port is supported. */
	( void ) xPort;
}
/*-----------------------------------------------------------*/

void vU2RXInterruptHandler( void )
{
/* Declared static to minimise stack use. */
static char cChar;
static portBASE_TYPE xHigherPriorityTaskWoken;

	xHigherPriorityTaskWoken = pdFALSE;

	/* Are any Rx interrupts pending? */
	if( IFS4bits.U2RXIF == 1 )
	{
         /* Receive all data from RX FIFO that avilable*/
         while( U2STAbits.URXDA == 1)
		{
			/* Retrieve the received character and place it in the queue of
			  received characters. */
			cChar = U2RXREG;
			xQueueSendFromISR( xRxedChars, &cChar, &xHigherPriorityTaskWoken ); 
 
			 /* clear any RX errors */
			if (U2STAbits.OERR) U2STAbits.OERR = 0;
			if (U2STAbits.FERR) U2STAbits.FERR = 0;
			if (U2STAbits.PERR) U2STAbits.PERR = 0;
		}
	}
	/* Clear the RX interrupt pending Flag */
	IFS4bits.U2RXIF = 0;
    
	/* If sending or receiving necessitates a context switch, then switch now. */
	portEND_SWITCHING_ISR( xHigherPriorityTaskWoken );
}


void vU2TXInterruptHandler( void )
{
/* Declared static to minimise stack use. */
static char cChar;
static portBASE_TYPE xHigherPriorityTaskWoken;

	xHigherPriorityTaskWoken = pdFALSE;

 /* Check if UART TX interrupts pending */
    if (  IFS4bits.U2TXIF == 1)
	{
		while( !( U2STAbits.UTXBF ) ) /* Continue until UART TX fifo is full */
			{
				if( xQueueReceiveFromISR( xCharsForTx, &cChar, &xHigherPriorityTaskWoken ) == pdTRUE )
					{
						/* Send the next character queued for Tx. */
						U2TXREG = cChar;
					}
					else
					{
						/* Queue empty, nothing to send. */
						/* Turn off UART TX interrupt */
						IEC4bits.U2TXIE = 0;
						xTxHasEnded = pdTRUE;
						break;
					}
			}
	}

	IFS4bits.U2TXIF = 0; // clear interrupt

	/* If sending or receiving necessitates a context switch, then switch now. */
	portEND_SWITCHING_ISR( xHigherPriorityTaskWoken );
}

void vU2FAULTInterruptHandler( void )
{
/* Declared static to minimise stack use. */
//static char cChar;
static portBASE_TYPE xHigherPriorityTaskWoken;

	xHigherPriorityTaskWoken = pdFALSE;

  /* Any Rx errors interrupts pending? */
	if( IFS4bits.U2EIF == 1)
        {
            /* We need to clear any Overrun errors if any, or it will stop UART */
			if (U2STAbits.OERR) U2STAbits.OERR = 0;
			if (U2STAbits.FERR) U2STAbits.FERR = 0;
			if (U2STAbits.PERR) U2STAbits.PERR = 0;    
        }
	IFS4bits.U2EIF = 0; // clear interrupt
	/* If sending or receiving necessitates a context switch, then switch now. */
	portEND_SWITCHING_ISR( xHigherPriorityTaskWoken );
}

/* These routines are NOT RTOS safe */
/*-----------------------------------------------------------*/
void uart2_puts(char *str) {
    while (*str != '\0') {
        while (U2STAbits.UTXBF)
            ; /* wait if the buffer is full */
        U2TXREG = *str++; /* transfer data byte to TX reg */
    }
}

/*-----------------------------------------------------------*/
void uart2_put(char *str, int len) {
    while (len-- > 0) {
        while (U2STAbits.UTXBF)
            ; /* wait if the buffer is full */
        U2TXREG = *str++; /* transfer data byte to TX reg */
    }
}

/*-----------------------------------------------------------*/
void uart2_putc(char c) {
    while (U2STAbits.UTXBF)
        ;
    U2TXREG = c;
}

/*-----------------------------------------------------------*/
char uart2_getc(void) {
    while (U2STAbits.URXDA == 0)
        ;
	return(U2RXREG);
}
/* ********************************************************* */



